import os
import uuid
import getpass
from six.moves.urllib.parse import urljoin
import webbrowser
import tempfile

from pybloqs.email import send_html_report
from pybloqs.htmlconv import htmlconv
from pybloqs.static import DependencyTracker, JScript, Css, script_inflate, script_block_core, register_interactive
from pybloqs.util import Cfg, cfg_to_css_string, str_base
from pybloqs.config import user_config

from pybloqs.html import root, append_to, render, js_elem, id_generator

try:
    from cStringIO import StringIO
except ImportError:
    from six import StringIO

# Valid page sizes with widths in mm
_page_width = {
    "A0": 841,
    "A1": 594,
    "A2": 420,
    "A3": 297,
    "A4": 210,
    "A5": 148,
    "A6": 105,
    "A7": 74,
    "A8": 52,
    "A9": 37,
    "B0": 1000,
    "B1": 707,
    "B2": 500,
    "B3": 353,
    "B4": 250,
    "B5": 176,
    "B6": 125,
    "B7": 88,
    "B8": 62,
    "B9": 33,
    "B10": 31,
    "C5E": 163,
    "Comm10E": 105,
    "DLE": 110,
    "Executive": 190.5,
    "Folio": 210,
    "Ledger": 431.8,
    "Legal": 215.9,
    "Letter": 215.9,
    "Tabloid": 279.4
}


default_css_main = Css(os.path.join("css", "pybloqs_default", "main"))
register_interactive(default_css_main)


class BaseBlock(object):
    """
    Base class for all blocks. Provides infrastructure for rendering the block
    in an IPython Notebook or saving it to disk in HTML, PDF, PNG or JPG format.
    """
    container_tag = "div"
    resource_deps = []

    def __init__(self, title=None, title_level=3, title_wrap=False,
                 width=None, height=None, inherit_cfg=True,
                 styles=None, classes=(), anchor=None, **kwargs):
        self._settings = Cfg(title=title,
                             title_level=title_level,
                             title_wrap=title_wrap,
                             cascading_cfg=Cfg(**kwargs).override(styles or Cfg()),
                             default_cfg=Cfg(),
                             inherit_cfg=inherit_cfg,
                             width=width,
                             height=height,
                             classes=["pybloqs"] + ([classes] if isinstance(classes, str) else list(classes)))
        # Anchor should not be inherited, so keep outside of Cfg
        self._anchor = anchor
        self._id = str(uuid.uuid4())

    def render_html(self, pretty=True, static_output=False, header_block=None, footer_block=None, pdf_page_size="A4"):
        """Returns html output of the block
        :param pretty: Toggles pretty printing of the resulting HTML. Not applicable for non-HTML output.
        :return html-code of the block
        """
        # Render the contents
        html = root("html", doctype="html")
        head = append_to(html, "head")
        body = append_to(html, "body")

        # Make sure that the main style sheet is always included
        resource_deps = DependencyTracker(default_css_main)
        if header_block is not None:
            header_block._write_block(body, Cfg(), id_generator(), resource_deps=resource_deps,
                                      static_output=static_output)

        self._write_block(body, Cfg(), id_generator(), resource_deps=resource_deps, static_output=static_output)

        if footer_block is not None:
            footer_block._write_block(body, Cfg(), id_generator(), resource_deps=resource_deps,
                                      static_output=static_output)

        script_inflate.write(head)
        script_block_core.write(head)

        if static_output:
            # Add the load wait poller if there are any JS resources
            js_elem(body, "var loadWaitPoller=runWaitPoller();")

        # Write out resources
        for res in resource_deps:
            res.write(head)

        # Render the whole document (the parent of the html tag)
        content = render(html.parent, pretty=pretty)
        return content

    def save(self, filename=None, fmt=None, toc=False, pdf_zoom=1, pdf_page_size="A4", pdf_auto_shrink=True,
             pretty=True, orientation='Portrait', header_block=None, header_spacing=5, footer_block=None,
             footer_spacing=5, java_script_delay=200, **kwargs):
        """
        Render and save the block. Depending on whether the filename or the format is
        provided, the content will either be written out to a file or returned as a string.

        :param filename: Format will be based on the file extension.
                         The following formats are supported:
                         - HTML
                         - PDF
                         - PNG
                         - JPG
        :param fmt: Specifies the format of a temporary output file. When supplied, the filename
                    parameter must be omitted.
        :param toc: Toggles the generation of Table of Contents.
                    Note: currently supported for PDF output only.
        :param pdf_zoom: The zooming to apply when rendering the page to PDF.
        :param pdf_page_size: The page size to use when rendering the page to PDF.
        :param pdf_auto_shrink: Toggles auto-shrinking content to fit the desired page size.
        :param pretty: Toggles pretty printing of the resulting HTML. Not applicable for non-HTML output.
        :return: html filename
        """
        # Ensure that exactly one of filename or fmt is provided
        if filename is None and fmt is None:
            raise ValueError("One of `filename` or `fmt` must be provided.")

        tempdir = tempfile.gettempdir()

        if filename:
            _, fmt_from_name = os.path.splitext(filename)
            # Exclude the dot from the extension, gosh darn it!
            fmt_from_name = fmt_from_name[1:]
            if fmt is None:
                if fmt_from_name == '':
                    raise ValueError('If fmt is not specified, filename must contain extension')
                fmt = fmt_from_name
            else:
                if fmt != fmt_from_name:
                    filename += '.' + fmt
        else:
            name = str_base(abs(hash(self._id))) + "." + fmt
            filename = os.path.join(tempdir, name)

        # Force extension to be lower case so format checks are easier later
        fmt = fmt.lower()

        is_html = "htm" in fmt

        if is_html:
            content = self.render_html(static_output=False, header_block=header_block, footer_block=footer_block)
            html_filename = filename
        else:
            content = self.render_html(static_output=True, pdf_page_size=pdf_page_size)
            name = str_base(abs(hash(self._id))) + ".html"
            html_filename = os.path.join(tempdir, name)

        # File with HTML content is needed either directly or as input for conversion
        with open(html_filename, "w") as f:
            f.write(content)

        if not is_html:

            cmd = ["--no-stop-slow-scripts", "--debug-javascript", "--javascript-delay", str(java_script_delay)]

            if fmt.endswith("pdf"):
                if pdf_page_size is not None:
                    cmd.extend(["--page-size", pdf_page_size])
                cmd.extend(["--orientation", orientation])
                if pdf_zoom != 1:
                    cmd.extend(["--zoom", str(pdf_zoom)])

                if pdf_auto_shrink:
                    cmd.append("--enable-smart-shrinking")
                else:
                    cmd.append("--disable-smart-shrinking")

                if header_block is not None:
                    header_file = str_base(hash(header_block._id)) + ".html"
                    file_path = header_block.publish(os.path.join(tempdir, header_file))
                    cmd += ['--header-html', file_path]
                    cmd += ['--header-spacing', str(header_spacing)]

                if footer_block is not None:
                    footer_file = str_base(hash(footer_block._id)) + ".html"
                    file_path = footer_block.publish(os.path.join(tempdir, footer_file))
                    cmd += ['--footer-html', file_path]
                    cmd += ['--footer-spacing', str(footer_spacing)]

            else:
                # In case the output is an image (not html or pdf), set width to 0
                # so that it fits the content exactly, without any extra margin.
                cmd.extend(["--width", "0"])
            # Add kwargs as additional htmlto... arguments
            [cmd.extend([k, v]) for k, v in kwargs.items()]

            htmlconv(input_file=html_filename, fmt=fmt, toc=toc, tool_args=cmd, output_file=filename)
            return filename

        return html_filename

    def publish(self, name, *args, **kwargs):
        """
        Publish the block so that others can access it.

        :param name: Name to publish under. Can be a filename or a relative path.
        :param args: Arguments to pass to `Block.save`.
        :param kwargs: Keyword arguments to pass to `Block.save`.
        :return: Path to the published block file.
        """
        full_path = os.path.join(user_config["public_dir"], name)
        full_path = os.path.expanduser(full_path)

        base_dir = os.path.dirname(full_path)

        try:
            os.makedirs(base_dir)
        except OSError:
            pass  # Directory already exists

        self.save(full_path, * args, **kwargs)

        return full_path

    def show(self, fmt="html", header_block=None, footer_block=None):
        """
        Show the block in a browser.

        :param fmt: The format of the saved block. Supports the same output as `Block.save`
        :return: Path to the block file.
        """
        file_name = str_base(hash(self._id)) + "." + fmt
        file_path = self.publish(os.path.expanduser(os.path.join(user_config["tmp_html_dir"], file_name)),
                                 header_block=header_block, footer_block=footer_block)

        try:
            url_base = user_config["public_dir"]
        except KeyError:
            path = os.path.expanduser(file_path)
        else:
            path = urljoin(url_base, os.path.expanduser(user_config["tmp_html_dir"] + "/" + file_name))

        webbrowser.open_new_tab(path)

        return path

    def email(self, title="", recipients=(user_config["user_email_address"],),
              footer_text=None, header_block=None, footer_block=None,
              from_address=None, cc=None, attachments=None, **kwargs):
        """
        Send the rendered blocks as email. Each output format chosen will be added as an
        attachment.

        :param title: title of the email
        :param recipients: recipient of the email
        :param fmt: One or more output formats that should be included as attachments.
                    The following formats are supported:
                    - HTML
                    - PDF
                    - PNG
                    - JPG
        :param body_block: The block to use as the email body. The default behavior is
                          to use the current block.
        :param footer_text: string to be used in place of the default footer text
        :param from_address: sender of the message. Defaults to user name. 
            Can be overwritten in .pybloqs.cfg with yaml format: 'user_email_address: a@b.com'
        :param cc: cc recipient
        :param kwargs: Optional arguments to pass to `Block.save()`
        """
        if from_address is None:
            from_address = user_config["user_email_address"]

        # The email body needs to be static without any dynamic elements.
        email_html = self.render_html(header_block=header_block, footer_block=footer_block)

        send_html_report(email_html, recipients, subject=title, attachments=attachments, From=from_address, Cc=cc)

    def to_static(self):
        return self._visit(lambda block: block._to_static())

    def _to_static(self):
        """
        Subclasses can override this method to provide a static content version.
        """
        return self

    def _visit(self, visitor):
        """
        Calls the supplied visitor function on this block and any sub-blocks
        :param visitor: Visitor function
        :return: Return value of the visitor
        """
        return visitor(self)

    def _provide_default_cfg(self, defaults):
        """
        Makes the supplied config to be part of the defaults for the block.
        :param defaults: The default parameters that should be inherited.
        """
        self._settings.default_cfg = self._settings.default_cfg.inherit(defaults)

    def _combine_parent_cfg(self, parent_cfg):
        """from pybloqs.config import user_config
        Combine the supplied parent and the current Block's config.

        :param parent_cfg: Parent config to inherit from.
        :return: Combined config.
        """
        # Combine parameters only if inheritance is turned on
        if self._settings.inherit_cfg:
            actual_cfg = self._settings.cascading_cfg.inherit(parent_cfg)
        else:
            actual_cfg = self._settings.cascading_cfg

        # Any undefined settings will use the defaults
        actual_cfg = actual_cfg.inherit(self._settings.default_cfg)

        return actual_cfg

    def _get_styles_string(self, styles_cfg):
        """
        Converts the styles configuration to a CSS styles string.

        :param styles_cfg: The configuration object to convert.
        :return: CSS string
        """
        sizing_cfg = Cfg()

        if self._settings.width is not None:
            sizing_cfg["width"] = self._settings.width

        if self._settings.height is not None:
            sizing_cfg["height"] = self._settings.height

        # Replace `_` with `-` and make values lowercase to get valid CSS names
        return cfg_to_css_string(styles_cfg.override(sizing_cfg))

    def _write_block(self, parent, parent_cfg, id_gen, resource_deps=None, static_output=False):
        """
        Writes out the block into the supplied stream, inheriting the parent_parameters.

        :param parent: Parent element
        :param parent_cfg: Parent parameters to inherit.
        :param id_gen: Unique ID generator.
        :param resource_deps: Object used to register resource dependencies.
        :param static_output: A value of True signals to blocks that the final output will
                              be a static format. Certain dynamic content will render with
                              alternate options.

        """
        if resource_deps is not None:
            for res in self.resource_deps:
                resource_deps.add(res)

        actual_cfg = self._combine_parent_cfg(parent_cfg)

        if self.container_tag is not None:
            container = append_to(parent, self.container_tag)
            self._write_container_attrs(container, actual_cfg)
        else:
            container = parent

        self._write_anchor(container)
        self._write_title(container)
        self._write_contents(container, actual_cfg, id_gen, resource_deps=resource_deps, static_output=static_output)

    def _write_container_attrs(self, container, actual_cfg):
        """
        Writes out the container attributes (styles, class, etc...).
        Note that this method will only be called if the container tag is not `None`.

        :param container: Container element.
        :param actual_cfg: Actual parameters to use.
        """
        styles = self._get_styles_string(actual_cfg)
        if len(styles) > 0:
            container["style"] = styles

        container["class"] = self._settings.classes

    def _write_title(self, container):
        """
        Write out the title (if there is any).

        :param container: Container element.
        """
        if self._settings.title is not None and (self._settings.title != ""):
            title = append_to(container, "H%s" % self._settings.title_level,
                              style="white-space: %s" % ("normal" if self._settings.title_wrap else "nowrap"))
            title.string = self._settings.title

    def _write_anchor(self, container):
        """
        Write HTML anchor for linking within page

        :param container: Container element.
        """
        if self._anchor is not None:
            append_to(container, "a", name=self._anchor)

    def _write_contents(self, container, actual_cfg, id_gen, resource_deps=None, static_output=None):
        """
        Write out the actual contents of the block. Deriving classes must override
        this method.

        :param container: Container element.
        :param actual_cfg: Actual parameters to use.
        :param id_gen: Unique ID generator.
        :param resource_deps: Object used to register resource dependencies.
        :param static_output: A value of True signals to blocks that the final output will
                              be a static format. Certain dynamic content will render with
                              alternate options.
        """
        raise NotImplementedError("_write_contents")

    def _repr_html_(self, *_):
        """
        Function required to support interactive IPython plopping and plotting.

        Should not be used directly.

        :return: Data to be displayed
        """
        return self.data

    @property
    def data(self):
        """
        Function required to support interactive IPython plotting.

        Should not be used directly.

        :return: Data to be displayed
        """
        container = root("div")
        self._write_block(container, Cfg(), id_generator())

        # Write children into the output
        output = StringIO()

        for child in container.children:
            output.write(render(child))

        return output.getvalue()


class HRule(BaseBlock):
    """
    Draws a horizontal divider line.
    """

    def _write_block(self, parent, *args, **kwargs):
        # Add a `hr` element to the parent
        append_to(parent, "hr")



import itertools

from pybloqs.html import js_elem, append_to, root, render
from pybloqs.util import encode_string

from pkg_resources import resource_filename

try:
    from cStringIO import StringIO
except ImportError:
    from six import StringIO


class Resource(object):
    def __init__(self, name, extension):
        """
        An external script dependency definition.

        :param name: Canonical name of the script.
        :param extension: Script file extension.
        """
        self._name = name
        self._extension = extension
        self._hash = hash(name)

    @property
    def _local_path(self):
        """
        Generate the local path to the script using pkg_resources and taking the minification
        switch into account.
        """
        return resource_filename(__name__, self._name + self._extension)

    def write(self, parent):
        pass

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if not isinstance(other, Resource):
            return False
        return self._hash == other._hash


class JScript(Resource):
    """
    Definition for external JS script dependencies.
    """
    # global flag to turn off script enconding/compression accross the board
    global_encode = True

    def __init__(self, name, encode=True):
        """
        An javascript dependency definition.

        :param name: Canonical name of the script
        :param encode: Whether to compress and base64 encode the script.
        """
        super(JScript, self).__init__(name, ".js")
        self._encode = encode
        self._sentinel_var_name = "_pybloqs_load_sentinel_%s" % name.replace("-", "_")

    def write(self, parent=None):
        stream = StringIO()

        # Write out init condition
        stream.write("if(typeof(%s) == 'undefined'){" % self._sentinel_var_name)

        with open(self._local_path, "rb") as f:
            if self._encode:
                self.write_compressed(stream, f.read())
            else:
                stream.write(f.read())

        stream.write("%s = true;" % self._sentinel_var_name)

        stream.write("}")

        return js_elem(parent, stream.getvalue())

    @classmethod
    def write_compressed(cls, stream, data):
        if cls.global_encode:
            stream.write('blocksEval(RawDeflate.inflate(atob("')
            stream.write(encode_string(data))
            stream.write('")));')
        else:
            stream.write(data)


class Css(Resource):
    def __init__(self, name, tag_id=None):
        super(Css, self).__init__(name, ".css")
        self._tag_id = tag_id

    def write(self, parent=None):
        if parent is None:
            el = root("style")
        else:
            el = append_to(parent, "style")

        el["type"] = "text/css"

        if self._tag_id is not None:
            el["id"] = self._tag_id

        with open(self._local_path, "rb") as f:
            el.string = f.read()

        return el


class DependencyTracker(object):
    def __init__(self, *args):
        self._deps = list(args)
        self._deps_set = set(args)

    def add(self, *resources):
        for resource in resources:
            if resource not in self._deps_set:
                self._deps.append(resource)
                self._deps_set.add(resource)

    def any(self, type_filter=None):
        if type_filter is None:
            return len(self._deps) > 0

        return any(itertools.imap(lambda res: isinstance(res, type_filter), self._deps))

    def __iter__(self):
        return iter(self._deps)


#JS deflation script and the reporting core functionality is always registered
script_block_core = JScript("block-core", encode=False)
script_inflate = JScript("jsinflate", encode=False)

_registered_resources = DependencyTracker(script_block_core, script_inflate)


def register_interactive(*scripts):
    _registered_resources.add(*scripts)


def write_interactive(stream):
    for script in _registered_resources:
        stream.write(render(script.write(), pretty=False))



import base64
import struct

import matplotlib.pyplot as plt

from matplotlib.artist import Artist
from matplotlib.figure import Figure

from pybloqs.html import append_to
from pybloqs.block.base import BaseBlock
from pybloqs.block.convenience import add_block_types
from pybloqs.util import cfg_to_css_string

try:
    from cStringIO import StringIO
except ImportError:
    from six import StringIO


_MIME_TYPES = {
    "png": "png",
    "svg": "svg+xml"
}

_PLOT_FORMAT = "png"
_PLOT_MIME_TYPE = _MIME_TYPES[_PLOT_FORMAT]
_PLOT_DPI = 100


def set_plot_format(plot_format=None, plot_dpi=None):
    global _PLOT_FORMAT
    global _PLOT_MIME_TYPE
    global _PLOT_DPI

    if plot_format is not None:
        _PLOT_FORMAT = plot_format
        _PLOT_MIME_TYPE = _MIME_TYPES[plot_format]

    if plot_dpi is not None:
        _PLOT_DPI = plot_dpi


class ImgBlock(BaseBlock):

    def __init__(self, data, mime_type="png", width=None, height=None, img_style=None, **kwargs):
        """
        Create a block containing an image. The dimensions can be sniffed from GIF
        and PNG data, for other formats, the `width` and `height` parameters can be
        used to specify the size.

        :param data: Either a block to convert to an image, or raw image data.
        :mime_type: The mime type of the image (e.g. "png" or "svg+xml")
        :param width: Image width override as string, e.g. "50px". If unspecified, the actual
                      image width is used.
        :param width: Image height override as string, e.g. "50px". If unspecified, the actual
                      image height is used.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        # Wrapping a block in an image will render it
        if hasattr(data, "_write_block"):
            img_file = data.save(fmt=mime_type)
            data = open(img_file, "rb").read()

        self._img_data = base64.b64encode(data)
        self._mime_type = mime_type

        if width is None and height is None:
            if mime_type.lower() == "png":
                assert struct.unpack('ccc', data[1:4]) == ('P', 'N', 'G')
                x, y = struct.unpack(">ii", data[16:24])
            elif mime_type.lower() == "gif":
                x, y = struct.unpack("<HH", data[6:10])
            else:
                raise ValueError("Can't determine image dimensions for mime type %s" % mime_type)

            width, height = ("%spx" % x, "%spx" % y)

        if img_style is None:
            img_styles = {}
        else:
            img_styles = img_style

        if width is not None:
            img_styles["width"] = width
        if height is not None:
            img_styles["height"] = height

        self._img_styles = img_styles

        super(ImgBlock, self).__init__(**kwargs)

    def _write_contents(self, container, *args, **kwargs):
        src = StringIO()
        src.write("data:image/{};base64,".format(self._mime_type))
        src.write(self._img_data)

        img = append_to(container, "img", src=src.getvalue())

        if len(self._img_styles) > 0:
            img["style"] = cfg_to_css_string(self._img_styles)

    @staticmethod
    def from_file(img_file, **kwargs):
        """
        Load an image block from a file.

        :param img_file: File path or file-like object.
        :param kwargs: Arguments to pass to the `ImgBlock` constructor.
        :return: ImgBlock instance.
        """
        close_file = False

        if isinstance(img_file, str):
            img_file = open(img_file, "rb")
            close_file = True

        try:
            return ImgBlock(img_file.read(), **kwargs)
        finally:
            # Close the file in case it was opened in this function
            if close_file:
                img_file.close()


class PlotBlock(ImgBlock):

    def __init__(self, plot, close_plot=True, bbox_inches="tight",
                 width=None, height=None, **kwargs):
        """
        Create a block containing a matplotlib figure

        :param plot: A matplotlib figure, axes or artist object.
        :close_plot: Optional (default=True). Set to True to close the plot after it is
                     captured into an image and avoid lingering plot windows.
        :bbox_inches: Optional bounding box parameter for 'figure.savefig'.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        assert isinstance(plot, Artist), "PlotBlock contents must be matplotlib Artists"

        if isinstance(plot, Figure):
            figure = plot
        elif isinstance(plot, Artist):
            figure = plot.get_figure()
        else:
            raise ValueError("Unexpected plot object type %s", type(plot))

        img_data = StringIO()

        legends = []

        for ax in figure.get_axes():
            legend = ax.get_legend()
            if legend is not None:
                # Patch Legend get_window_extent since there seems to be a bug where
                # it is passed an unexpected renderer instance.
                _orig_get_window_extent = legend.get_window_extent

                def _patched_get_window_extent(*_):
                    return _orig_get_window_extent()

                legend.get_window_extent = _patched_get_window_extent
                legends.append(legend)

        if len(figure.axes) == 0:
            # empty plot, disable bbox_inches to that savefig still works
            bbox_inches = None

        figure.savefig(img_data, dpi=_PLOT_DPI, format=_PLOT_FORMAT,
                       bbox_extra_artists=legends, bbox_inches=bbox_inches)

        plt_width, plt_height = figure.get_size_inches()

        width = width or "{:0.3f}in".format(plt_width)
        height = height or "{:0.3f}in".format(plt_height)

        if close_plot:
            plt.close(figure)

        super(PlotBlock, self).__init__(img_data.getvalue(),
                                        _PLOT_MIME_TYPE,
                                        width=width,
                                        height=height,
                                        **kwargs)

    def _to_static(self):
        # Convert to a basic image block in case we contain 'dynamic' svg content
        return ImgBlock(self) if self._mime_type == "svg" else super(PlotBlock, self)._to_static()


add_block_types(Artist, PlotBlock)



import os
import subprocess
import sys


_CONF_PATH = os.path.join(sys.exec_prefix, "conf")


def htmlconv(html_string=None, fmt="pdf", toc=True, tool_args=(), input_file=None, output_file=None):
    """
    Use the `wkhtmlto*` tool to convert the supplied html content to a PDF or image file.

    :param html_string: Html string.
    :param fmt: Format to use. Can be one of "pdf", "jpg", "png".
    :param tool_args: Optional. Extra command line args to pass to the converter tool.
    :return: Res
    """
    if not ((html_string is None) ^ (input_file is None)):
        raise ValueError("Exactly one of `html_string` or `input_file` must be provided.")

    py_bin_dir = os.path.join(sys.exec_prefix, 'bin')

    fmt = fmt.lower()
    if fmt == "pdf":
        local_bin = os.path.join(py_bin_dir, 'wkhtmltopdf')
        if os.path.isfile(local_bin):
            cmd = [local_bin]
        else:
            cmd = ["wkhtmltopdf"]
        cmd += list(tool_args)
        if toc:
            cmd.extend(["toc", "--xsl-style-sheet", os.path.join(_CONF_PATH, "toc.xsl")])
    else:
        local_bin = os.path.join(py_bin_dir, 'wkhtmltoimage')
        if os.path.isfile(local_bin):
            cmd = [local_bin]
        else:
            cmd = ["wkhtmltoimage"]
        cmd += ["--format", fmt] + list(tool_args)

    cmd.extend([input_file, output_file])

    proc = subprocess.Popen(cmd)
    # Wait for the process to exit
    output, errors = proc.communicate()

    msg = "%s returned:\n\n%s" % (cmd, errors)
    if proc.returncode != 0:
        raise ValueError(msg)
    else:
        print(msg)

    # Return the raw output if no output file was provided
    return output, errors



from collections import namedtuple
import itertools
from jinja2 import Environment, PackageLoader

from pybloqs.block.base import BaseBlock
from pybloqs.block.convenience import add_block_types
from pybloqs.block.table_formatters import DEFAULT_FORMATTERS, DEFAULT_DECIMALS_FORMATTER, ORG_ROW_NAMES
from pybloqs.html import parse

import pandas as pd

_jinja_env = Environment(loader=PackageLoader('pybloqs', 'jinja'))
_jinja_env.globals.update(len=len)
_jinja_env.globals.update(enumerate=enumerate)
_table_tmpl = _jinja_env.get_template('table.html')


class HTMLJinjaTableBlock(BaseBlock):
    FormatterData = namedtuple('FormatterData', ['cell', 'row_name', 'column_name', 'df'])

    def __init__(self, df, formatters=None, use_default_formatters=True, **kwargs):
        """Create table from Jinga framework. Apply formatters to customise table formatting. 


        Parameters()
        ----------
        df: 'DataFrame'
            The DF from which the values are taken.
        formatters: 'list'
            List of formatters, which are objects inheriting from TableFormatter class.
        """
        super(HTMLJinjaTableBlock, self).__init__(**kwargs)
        if formatters is None:
            formatters = []
        if use_default_formatters:
            formatters = DEFAULT_FORMATTERS + formatters + DEFAULT_DECIMALS_FORMATTER

        self.formatters = formatters
        # Apply modifications to DataFrame at the earliest stage.
        for formatter in formatters:
            try:
                df = formatter.modify_dataframe(df)
            except NotImplementedError:
                continue
        self.df = df
        self.n_header_rows = len(df.columns.names)
        return

    def modify_cell_content(self, cell, row_name, column_name):
        if ORG_ROW_NAMES in self.df.columns and self.row_index > 0:
            row_name = self.df[ORG_ROW_NAMES].iloc[self.row_index]
        for formatter in self.formatters:
            try:
                data = self.FormatterData(cell, row_name, column_name, self.df)
                cell = formatter.modify_cell_content(data)
            except NotImplementedError:
                continue
        return cell

    def insert_additional_html(self):
        html_string = ''
        for formatter in self.formatters:
            try:
                html_string += formatter.insert_additional_html()
            except NotImplementedError:
                continue
        return html_string

    def _join_css_substrings(self, css_substrings, prefix):
        return prefix + "=\"" + "; ".join(css_substrings) + "\""

    def _aggregate_css_formatters(self, function_name, fmt_args=None, prefix='style'):
        css_substrings = []
        fmt_args = fmt_args if fmt_args else []
        for formatter in self.formatters:
            try:
                fmt_func = getattr(formatter, function_name)
                css_substring = fmt_func(*fmt_args)
            except NotImplementedError:
                continue
            if not css_substring is None:
                css_substrings.append(css_substring)
        return self._join_css_substrings(css_substrings, prefix)

    def create_table_level_css(self):
        self.row_index = -self.n_header_rows - 1
        return self._aggregate_css_formatters('create_table_level_css')
    
    def create_table_level_css_class(self):
        return self._aggregate_css_formatters('create_table_level_css_class', prefix='class')

    def create_thead_level_css(self):
        return self._aggregate_css_formatters('create_thead_level_css')

    def create_row_level_css(self, row_name, row):
        self.row_index += 1
        if ORG_ROW_NAMES in self.df.columns and self.row_index >= 0:
            row_name = self.df[ORG_ROW_NAMES].iloc[self.row_index]
        data = pd.Series(row, name=row_name)
        return self._aggregate_css_formatters('create_row_level_css', fmt_args=[data])

    def create_cell_level_css(self, cell, row_name, column_name):
        if ORG_ROW_NAMES in self.df.columns and self.row_index >= 0:
            row_name = self.df[ORG_ROW_NAMES].iloc[self.row_index]
        data = self.FormatterData(cell, row_name, column_name, self.df)
        return self._aggregate_css_formatters('create_cell_level_css', fmt_args=[data])
    
    def _get_header_iterable(self):
        """Reformats all but the last header rows."""
        df_clean = self.df.loc[:, self.df.columns.get_level_values(0) != ORG_ROW_NAMES]
        if isinstance(df_clean.columns, pd.MultiIndex):
            transpose_tuples = zip(*df_clean.columns.tolist())
            header_values = []
            for i, t in enumerate(transpose_tuples):
                if i < len(transpose_tuples) - 1:
                    # Not the last column, aggregate repeated items, e.g. [['aa', 'aa', 'aa'], ['bb', 'bb', 'bb']]
                    header_values.append([list(g) for _, g in itertools.groupby(t)])
                else:
                    # For the last column keep all elements in single list, e.g. ['a', 'b', 'c', 'a', 'b', 'c']
                    header_values.append(list(t))
            return header_values
        else:
            return [df_clean.columns.tolist()]

    def _write_contents(self, container, actual_cfg, *args, **kwargs):
        # table boilerplate
        model = {'df': self.df,
                 'header_iterable': self._get_header_iterable(),
                 'insert_additional_html': self.insert_additional_html,
                 'create_thead_level_css': self.create_thead_level_css,
                 'create_table_level_css': self.create_table_level_css,
                 'create_table_level_css_class' : self.create_table_level_css_class,
                 'create_row_level_css': self.create_row_level_css,
                 'create_cell_level_css': self.create_cell_level_css,
                 'modify_cell_content': self.modify_cell_content}

        table_html = _table_tmpl.render(**model)
        soup = parse(table_html)
        table = soup.find("table")
        container.append(table)
        return

add_block_types(pd.DataFrame, HTMLJinjaTableBlock)



from pybloqs.util import Cfg

from pybloqs.block.base import BaseBlock, HRule
from pybloqs.block.text import Raw, Pre, Span, Markdown
from pybloqs.block.layout import Grid, Flow, HStack, VStack
from pybloqs.block.convenience import Block
from pybloqs.block.image import ImgBlock, PlotBlock, set_plot_format
from pybloqs.block.table import HTMLJinjaTableBlock
from pybloqs.block.wrap import Box, Paragraph
from pybloqs import plot as bxpl

try:
    from cStringIO import StringIO
except ImportError:
    from six import StringIO


def interactive(verbose=True):
    """
    Enables interactive usage of block layout content.
    """
    from IPython.core.display import display_html
    from pybloqs.static import write_interactive
    from pybloqs.html import set_id_generator, id_generator_uuid

    set_id_generator(id_generator_uuid)

    stream = StringIO()

    write_interactive(stream)

    if verbose:
        stream.write("<div>Interactive mode initialized successfully</div>")

    # Send the scripts to the frontend
    return display_html(stream.getvalue(), raw=True)


__all__ = [
    # Core PyBloqs
    "set_plot_format",
    "Block",
    "BaseBlock",
    "HRule",
    "Raw",
    "Grid",
    "Flow",
    "HStack",
    "VStack",
    "ImgBlock",
    "PlotBlock",
    "Box",
    "Paragraph",
    "Pre",
    "Span",
    "Markdown",
    "Cfg",

    # Interactive Plots
    "interactive",
    "bxpl",
]



import bs4
import uuid

# Use the default python parser as this is lenient and does not
# wrap content in extra tags
PYBLOQS_ID_PREFIX = "pybloqs_id_"
DEFAULT_PARSER = "html.parser"


def parse(string):
    """
    Parses the string into a beautiful soup tree.

    :param string: String to parse.
    :return: Soup.
    """
    return bs4.BeautifulSoup(string, DEFAULT_PARSER)


def root(tag_name="html", doctype=None):
    """
    Creates a new soup with the given root element.

    :param tag_name: Root element tag name.
    :param doctype: Optional doctype tag to add.
    :return: Soup.
    """
    soup = parse("")
    if doctype is not None:
        soup.append(bs4.Doctype(doctype))
    tag = soup.new_tag(tag_name)
    tag.soup = soup
    soup.append(tag)
    return tag


def render(item, pretty=True, encoding="utf8"):
    """
    Renders the given element into a string.

    :param item: Item to render. Must be an element or soup instance.
    :param pretty: Toggles pretty formatting of the resulting string.
    :return: Rendered content.
    """
    return item.prettify(encoding=encoding) if pretty else str(item)


def append_to(parent, tag, **kwargs):
    """
    Append an element to the supplied parent.

    :param parent: Parent to append to.
    :param tag: Tag to create.
    :param args: Tag args.
    :param kwargs: Tag kwargs.
    :return: New element.
    """
    if hasattr(parent, "soup"):
        soup = parent.soup
    else:
        soup = parent.find_parent("html")

    # Create Tag explicitly instead of using new_tag, otherwise attribute "name" leads to clash with tag-name in bs4
    new_tag = bs4.Tag(builder=soup.builder, name=tag, attrs=kwargs)

    new_tag.soup = soup

    parent.append(new_tag)

    return new_tag


def js_elem(container=None, script=None):
    """
    Constructs a Javascript element and appends it to the container.

    :param container: Container to add the element to.
    :param script: Javascript code.
    :return: New element.
    """
    if container is None:
        el = root("script")
    else:
        el = append_to(container, "script", type="text/javascript")
    if script is not None:
        el.string = script
    return el


def set_id_generator(generator):
    """
    Sets the global id generator function (must be a python generator function)

    :param generator: Generator function to use.
    """
    global _id_generator
    _id_generator = generator


def id_generator_uuid():
    """
    Generates unique identifiers using the `uuid` package.
    """
    while True:
        yield str(uuid.uuid4()).replace("-", "")


def id_generator_sequential():
    """
    Generatres unique identifiers sequentially from a known constant seed.
    """
    counter = 0

    while True:
        yield PYBLOQS_ID_PREFIX + str(counter)
        counter += 1


def id_generator():
    return _id_generator()

_id_generator = id_generator_sequential



import os
import yaml
import getpass

# Default values that can be overridden by those stored in ~/.pybloqs.cfg
user_config = {
    "user_email_address": getpass.getuser(),
    "public_dir": "/tmp",
    "tmp_html_dir": "/tmp",
    "smtp_server": ""
}

try:
    stored_config = yaml.load(open(os.path.expanduser("~/.pybloqs.cfg"), "r"))
    if stored_config is not None:
        user_config.update(stored_config)
except IOError:
    pass
